# Copyright 2018 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import importlib.util
import os
import sys
import unittest

def _GetDirAbove(dirname):
  """Returns the directory "above" this file containing |dirname| (which must
  also be "above" this file)."""
  path = os.path.abspath(__file__)
  while True:
    path, tail = os.path.split(path)
    assert tail
    if tail == dirname:
      return path

try:
  importlib.util.find_spec("mojom")
except ImportError:
  sys.path.append(os.path.join(_GetDirAbove('pylib'), 'pylib'))
import mojom.parse.ast as ast
import mojom.parse.conditional_features as conditional_features
import mojom.parse.parser as parser

ENABLED_FEATURES = frozenset({'red', 'green', 'blue'})

class ConditionalFeaturesTest(unittest.TestCase):
  """Tests |mojom.parse.conditional_features|."""

  def parseAndAssertEqual(self, source, expected_source):
    definition = parser.Parse(source, "my_file.mojom")
    conditional_features.RemoveDisabledDefinitions(definition, ENABLED_FEATURES)
    expected = parser.Parse(expected_source, "my_file.mojom")
    self.assertEquals(definition, expected)

  def testFilterConst(self):
    """Test that Consts are correctly filtered."""
    const_source = """
      [EnableIf=blue]
      const int kMyConst1 = 1;
      [EnableIf=orange]
      const double kMyConst2 = 2;
      const int kMyConst3 = 3;
    """
    expected_source = """
      [EnableIf=blue]
      const int kMyConst1 = 1;
      const int kMyConst3 = 3;
    """
    self.parseAndAssertEqual(const_source, expected_source)

  def testFilterIfNotConst(self):
    """Test that Consts are correctly filtered."""
    const_source = """
      [EnableIfNot=blue]
      const int kMyConst1 = 1;
      [EnableIfNot=orange]
      const double kMyConst2 = 2;
      [EnableIf=blue]
      const int kMyConst3 = 3;
      [EnableIfNot=blue]
      const int kMyConst4 = 4;
      [EnableIfNot=purple]
      const int kMyConst5 = 5;
    """
    expected_source = """
      [EnableIfNot=orange]
      const double kMyConst2 = 2;
      [EnableIf=blue]
      const int kMyConst3 = 3;
      [EnableIfNot=purple]
      const int kMyConst5 = 5;
    """
    self.parseAndAssertEqual(const_source, expected_source)

  def testFilterIfNotMultipleConst(self):
    """Test that Consts are correctly filtered."""
    const_source = """
      [EnableIfNot=blue]
      const int kMyConst1 = 1;
      [EnableIfNot=orange]
      const double kMyConst2 = 2;
      [EnableIfNot=orange]
      const int kMyConst3 = 3;
    """
    expected_source = """
      [EnableIfNot=orange]
      const double kMyConst2 = 2;
      [EnableIfNot=orange]
      const int kMyConst3 = 3;
    """
    self.parseAndAssertEqual(const_source, expected_source)

  def testFilterEnum(self):
    """Test that EnumValues are correctly filtered from an Enum."""
    enum_source = """
      enum MyEnum {
        [EnableIf=purple]
        VALUE1,
        [EnableIf=blue]
        VALUE2,
        VALUE3,
      };
    """
    expected_source = """
      enum MyEnum {
        [EnableIf=blue]
        VALUE2,
        VALUE3
      };
    """
    self.parseAndAssertEqual(enum_source, expected_source)

  def testFilterImport(self):
    """Test that imports are correctly filtered from a Mojom."""
    import_source = """
      [EnableIf=blue]
      import "foo.mojom";
      import "bar.mojom";
      [EnableIf=purple]
      import "baz.mojom";
    """
    expected_source = """
      [EnableIf=blue]
      import "foo.mojom";
      import "bar.mojom";
    """
    self.parseAndAssertEqual(import_source, expected_source)

  def testFilterIfNotImport(self):
    """Test that imports are correctly filtered from a Mojom."""
    import_source = """
      [EnableIf=blue]
      import "foo.mojom";
      [EnableIfNot=purple]
      import "bar.mojom";
      [EnableIfNot=green]
      import "baz.mojom";
    """
    expected_source = """
      [EnableIf=blue]
      import "foo.mojom";
      [EnableIfNot=purple]
      import "bar.mojom";
    """
    self.parseAndAssertEqual(import_source, expected_source)

  def testFilterInterface(self):
    """Test that definitions are correctly filtered from an Interface."""
    interface_source = """
      interface MyInterface {
        [EnableIf=blue]
        enum MyEnum {
          [EnableIf=purple]
          VALUE1,
          VALUE2,
        };
        [EnableIf=blue]
        const int32 kMyConst = 123;
        [EnableIf=purple]
        MyMethod();
      };
    """
    expected_source = """
      interface MyInterface {
        [EnableIf=blue]
        enum MyEnum {
          VALUE2,
        };
        [EnableIf=blue]
        const int32 kMyConst = 123;
      };
    """
    self.parseAndAssertEqual(interface_source, expected_source)

  def testFilterMethod(self):
    """Test that Parameters are correctly filtered from a Method."""
    method_source = """
      interface MyInterface {
        [EnableIf=blue]
        MyMethod([EnableIf=purple] int32 a) => ([EnableIf=red] int32 b);
      };
    """
    expected_source = """
      interface MyInterface {
        [EnableIf=blue]
        MyMethod() => ([EnableIf=red] int32 b);
      };
    """
    self.parseAndAssertEqual(method_source, expected_source)

  def testFilterStruct(self):
    """Test that definitions are correctly filtered from a Struct."""
    struct_source = """
      struct MyStruct {
        [EnableIf=blue]
        enum MyEnum {
          VALUE1,
          [EnableIf=purple]
          VALUE2,
        };
        [EnableIf=yellow]
        const double kMyConst = 1.23;
        [EnableIf=green]
        int32 a;
        double b;
        [EnableIf=purple]
        int32 c;
        [EnableIf=blue]
        double d;
        int32 e;
        [EnableIf=orange]
        double f;
      };
    """
    expected_source = """
      struct MyStruct {
        [EnableIf=blue]
        enum MyEnum {
          VALUE1,
        };
        [EnableIf=green]
        int32 a;
        double b;
        [EnableIf=blue]
        double d;
        int32 e;
      };
    """
    self.parseAndAssertEqual(struct_source, expected_source)

  def testFilterIfNotStruct(self):
    """Test that definitions are correctly filtered from a Struct."""
    struct_source = """
      struct MyStruct {
        [EnableIf=blue]
        enum MyEnum {
          VALUE1,
          [EnableIfNot=red]
          VALUE2,
        };
        [EnableIfNot=yellow]
        const double kMyConst = 1.23;
        [EnableIf=green]
        int32 a;
        double b;
        [EnableIfNot=purple]
        int32 c;
        [EnableIf=blue]
        double d;
        int32 e;
        [EnableIfNot=red]
        double f;
      };
    """
    expected_source = """
      struct MyStruct {
        [EnableIf=blue]
        enum MyEnum {
          VALUE1,
        };
        [EnableIfNot=yellow]
        const double kMyConst = 1.23;
        [EnableIf=green]
        int32 a;
        double b;
        [EnableIfNot=purple]
        int32 c;
        [EnableIf=blue]
        double d;
        int32 e;
      };
    """
    self.parseAndAssertEqual(struct_source, expected_source)

  def testFilterUnion(self):
    """Test that UnionFields are correctly filtered from a Union."""
    union_source = """
      union MyUnion {
        [EnableIf=yellow]
        int32 a;
        [EnableIf=red]
        bool b;
      };
    """
    expected_source = """
      union MyUnion {
        [EnableIf=red]
        bool b;
      };
    """
    self.parseAndAssertEqual(union_source, expected_source)

  def testSameNameFields(self):
    mojom_source = """
      enum Foo {
        [EnableIf=red]
        VALUE1 = 5,
        [EnableIf=yellow]
        VALUE1 = 6,
      };
      [EnableIf=red]
      const double kMyConst = 1.23;
      [EnableIf=yellow]
      const double kMyConst = 4.56;
    """
    expected_source = """
      enum Foo {
        [EnableIf=red]
        VALUE1 = 5,
      };
      [EnableIf=red]
      const double kMyConst = 1.23;
    """
    self.parseAndAssertEqual(mojom_source, expected_source)

  def testFeaturesWithEnableIf(self):
    mojom_source = """
      feature Foo {
        const string name = "FooFeature";
        [EnableIf=red]
        const bool default_state = false;
        [EnableIf=yellow]
        const bool default_state = true;
      };
    """
    expected_source = """
      feature Foo {
        const string name = "FooFeature";
        [EnableIf=red]
        const bool default_state = false;
      };
    """
    self.parseAndAssertEqual(mojom_source, expected_source)

  def testMultipleEnableIfs(self):
    source = """
      enum Foo {
        [EnableIf=red,EnableIf=yellow]
        kBarValue = 5,
      };
    """
    definition = parser.Parse(source, "my_file.mojom")
    self.assertRaises(conditional_features.EnableIfError,
                      conditional_features.RemoveDisabledDefinitions,
                      definition, ENABLED_FEATURES)

  def testMultipleEnableIfs(self):
    source = """
      enum Foo {
        [EnableIf=red,EnableIfNot=yellow]
        kBarValue = 5,
      };
    """
    definition = parser.Parse(source, "my_file.mojom")
    self.assertRaises(conditional_features.EnableIfError,
                      conditional_features.RemoveDisabledDefinitions,
                      definition, ENABLED_FEATURES)

  def testMultipleEnableIfs(self):
    source = """
      enum Foo {
        [EnableIfNot=red,EnableIfNot=yellow]
        kBarValue = 5,
      };
    """
    definition = parser.Parse(source, "my_file.mojom")
    self.assertRaises(conditional_features.EnableIfError,
                      conditional_features.RemoveDisabledDefinitions,
                      definition, ENABLED_FEATURES)

if __name__ == '__main__':
  unittest.main()



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Ideas On Board
#
# AWB module for tuning rkisp1

from .awb import AWB

import libtuning as lt


class AWBRkISP1(AWB):
    hr_name = 'AWB (RkISP1)'
    out_name = 'Awb'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def validate_config(self, config: dict) -> bool:
        return True

    def process(self, config: dict, images: list, outputs: dict) -> dict:
        output = {}

        output['colourGains'] = self.do_calculation(images)

        return output



# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>
#
# Debayering code from PiCamera documentation

from numpy.lib.stride_tricks import as_strided
import libcamera as libcam
import libcamera.utils
import numpy as np


def demosaic(data, r0, g0, g1, b0):
    # Separate the components from the Bayer data to RGB planes

    rgb = np.zeros(data.shape + (3,), dtype=data.dtype)
    rgb[r0[1]::2, r0[0]::2, 0] = data[r0[1]::2, r0[0]::2]  # Red
    rgb[g0[1]::2, g0[0]::2, 1] = data[g0[1]::2, g0[0]::2]  # Green
    rgb[g1[1]::2, g1[0]::2, 1] = data[g1[1]::2, g1[0]::2]  # Green
    rgb[b0[1]::2, b0[0]::2, 2] = data[b0[1]::2, b0[0]::2]  # Blue

    # Below we present a fairly naive de-mosaic method that simply
    # calculates the weighted average of a pixel based on the pixels
    # surrounding it. The weighting is provided by a byte representation of
    # the Bayer filter which we construct first:

    bayer = np.zeros(rgb.shape, dtype=np.uint8)
    bayer[r0[1]::2, r0[0]::2, 0] = 1  # Red
    bayer[g0[1]::2, g0[0]::2, 1] = 1  # Green
    bayer[g1[1]::2, g1[0]::2, 1] = 1  # Green
    bayer[b0[1]::2, b0[0]::2, 2] = 1  # Blue

    # Allocate an array to hold our output with the same shape as the input
    # data. After this we define the size of window that will be used to
    # calculate each weighted average (3x3). Then we pad out the rgb and
    # bayer arrays, adding blank pixels at their edges to compensate for the
    # size of the window when calculating averages for edge pixels.

    output = np.empty(rgb.shape, dtype=rgb.dtype)
    window = (3, 3)
    borders = (window[0] - 1, window[1] - 1)
    border = (borders[0] // 2, borders[1] // 2)

    rgb = np.pad(rgb, [
        (border[0], border[0]),
        (border[1], border[1]),
        (0, 0),
    ], 'constant')
    bayer = np.pad(bayer, [
        (border[0], border[0]),
        (border[1], border[1]),
        (0, 0),
    ], 'constant')

    # For each plane in the RGB data, we use a nifty numpy trick
    # (as_strided) to construct a view over the plane of 3x3 matrices. We do
    # the same for the bayer array, then use Einstein summation on each
    # (np.sum is simpler, but copies the data so it's slower), and divide
    # the results to get our weighted average:

    for plane in range(3):
        p = rgb[..., plane]
        b = bayer[..., plane]
        pview = as_strided(p, shape=(
            p.shape[0] - borders[0],
            p.shape[1] - borders[1]) + window, strides=p.strides * 2)
        bview = as_strided(b, shape=(
            b.shape[0] - borders[0],
            b.shape[1] - borders[1]) + window, strides=b.strides * 2)
        psum = np.einsum('ijkl->ij', pview)
        bsum = np.einsum('ijkl->ij', bview)
        output[..., plane] = psum // bsum

    return output


def to_rgb(fmt, size, data):
    w = size.width
    h = size.height

    if fmt == libcam.formats.YUYV:
        # YUV422
        yuyv = data.reshape((h, w // 2 * 4))

        # YUV444
        yuv = np.empty((h, w, 3), dtype=np.uint8)
        yuv[:, :, 0] = yuyv[:, 0::2]                    # Y
        yuv[:, :, 1] = yuyv[:, 1::4].repeat(2, axis=1)  # U
        yuv[:, :, 2] = yuyv[:, 3::4].repeat(2, axis=1)  # V

        m = np.array([
            [1.0, 1.0, 1.0],
            [-0.000007154783816076815, -0.3441331386566162, 1.7720025777816772],
            [1.4019975662231445, -0.7141380310058594, 0.00001542569043522235]
        ])

        rgb = np.dot(yuv, m)
        rgb[:, :, 0] -= 179.45477266423404
        rgb[:, :, 1] += 135.45870971679688
        rgb[:, :, 2] -= 226.8183044444304
        rgb = rgb.astype(np.uint8)

    elif fmt == libcam.formats.RGB888:
        rgb = data.reshape((h, w, 3))
        rgb[:, :, [0, 1, 2]] = rgb[:, :, [2, 1, 0]]

    elif fmt == libcam.formats.BGR888:
        rgb = data.reshape((h, w, 3))

    elif fmt in [libcam.formats.ARGB8888, libcam.formats.XRGB8888]:
        rgb = data.reshape((h, w, 4))
        rgb = np.flip(rgb, axis=2)
        # drop alpha component
        rgb = np.delete(rgb, np.s_[0::4], axis=2)

    elif str(fmt).startswith('S'):
        fmt = str(fmt)
        bayer_pattern = fmt[1:5]
        bitspp = int(fmt[5:])

        if bitspp == 8:
            data = data.reshape((h, w))
            data = data.astype(np.uint16)
        elif bitspp in [10, 12]:
            data = data.view(np.uint16)
            data = data.reshape((h, w))
        else:
            raise Exception('Bad bitspp:' + str(bitspp))

        idx = bayer_pattern.find('R')
        assert(idx != -1)
        r0 = (idx % 2, idx // 2)

        idx = bayer_pattern.find('G')
        assert(idx != -1)
        g0 = (idx % 2, idx // 2)

        idx = bayer_pattern.find('G', idx + 1)
        assert(idx != -1)
        g1 = (idx % 2, idx // 2)

        idx = bayer_pattern.find('B')
        assert(idx != -1)
        b0 = (idx % 2, idx // 2)

        rgb = demosaic(data, r0, g0, g1, b0)
        rgb = (rgb >> (bitspp - 8)).astype(np.uint8)

    else:
        rgb = None

    return rgb


# A naive format conversion to 24-bit RGB
def mfb_to_rgb(mfb: libcamera.utils.MappedFrameBuffer, cfg: libcam.StreamConfiguration):
    data = np.array(mfb.planes[0], dtype=np.uint8)
    rgb = to_rgb(cfg.pixel_format, cfg.size, data)
    return rgb



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
# Copyright (C) 2024, Paul Elder <paul.elder@ideasonboard.com>
#
# rkisp1.py - AGC module for tuning rkisp1

from .agc import AGC

import libtuning as lt


class AGCRkISP1(AGC):
    hr_name = 'AGC (RkISP1)'
    out_name = 'Agc'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # We don't actually need anything from the config file
    def validate_config(self, config: dict) -> bool:
        return True

    def _generate_metering_modes(self) -> dict:
        centre_weighted = [
            0, 0,  0, 0, 0,
            0, 6,  8, 6, 0,
            0, 8, 16, 8, 0,
            0, 6,  8, 6, 0,
            0, 0,  0, 0, 0
        ]

        spot = [
            0, 0,  0, 0, 0,
            0, 2,  4, 2, 0,
            0, 4, 16, 4, 0,
            0, 2,  4, 2, 0,
            0, 0,  0, 0, 0
        ]

        matrix = [1 for i in range(0, 25)]

        return {
            'MeteringCentreWeighted': centre_weighted,
            'MeteringSpot': spot,
            'MeteringMatrix': matrix
        }

    def _generate_exposure_modes(self) -> dict:
        normal = {'exposureTime': [100, 10000, 30000, 60000, 120000],
                  'gain': [2.0, 4.0, 6.0, 6.0, 6.0]}
        short = {'exposureTime': [100, 5000, 10000, 20000, 120000],
                 'gain': [2.0, 4.0, 6.0, 6.0, 6.0]}

        return {'ExposureNormal': normal, 'ExposureShort': short}

    def _generate_constraint_modes(self) -> dict:
        normal = {'lower': {'qLo': 0.98, 'qHi': 1.0, 'yTarget': 0.5}}
        highlight = {
            'lower': {'qLo': 0.98, 'qHi': 1.0, 'yTarget': 0.5},
            'upper': {'qLo': 0.98, 'qHi': 1.0, 'yTarget': 0.8}
        }

        return {'ConstraintNormal': normal, 'ConstraintHighlight': highlight}

    def _generate_y_target(self) -> list:
        return 0.5

    def process(self, config: dict, images: list, outputs: dict) -> dict:
        output = {}

        output['AeMeteringMode'] = self._generate_metering_modes()
        output['AeExposureMode'] = self._generate_exposure_modes()
        output['AeConstraintMode'] = self._generate_constraint_modes()
        output['relativeLuminanceTarget'] = self._generate_y_target()

        # \todo Debug functionality

        return output



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (C) 2020, Google Inc.
#
# Author: Paul Elder <paul.elder@ideasonboard.com>
#
# Run mojo code generator for generating libcamera IPC files

import os
import sys

sys.path.insert(0, f'{os.path.dirname(__file__)}/mojo/public/tools/bindings')

import mojo.public.tools.bindings.mojom_bindings_generator as generator

def _GetModulePath(path, output_dir):
    return os.path.join(output_dir, path.relative_path())


# Disable the attribute checker to support our custom attributes. Ideally we
# should add the attributes to the list of allowed attributes in
# utils/ipc/mojo/public/tools/bindings/checks/mojom_attributes_check.py, but
# we're trying hard to use the upstream mojom as-is.
if hasattr(generator, '_BUILTIN_CHECKS'):
    del generator._BUILTIN_CHECKS['attributes']

# Override the mojo code generator's generator list to only contain our
# libcamera generator
generator._BUILTIN_GENERATORS = {'libcamera': 'mojom_libcamera_generator'}

# Override the mojo code generator's _GetModulePath method to not add
# the '-module' suffix when searching for mojo modules, so that we can
# pass the path to the mojom module without having to trim the '-module' suffix
generator._GetModulePath = _GetModulePath

generator.main()
